/* zvt_access-- accessibility support for libzvt
 * Copyright 2001 Sun Microsystems Inc.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <stdio.h>
#include <errno.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <atk/atk.h>
#include <gtk/gtk.h>
#include <libzvt/libzvt.h>
#include "zvt_accessible.h"
#include "zvt_accessiblefactory.h"


typedef enum {
  DIRECTION_AT,
  DIRECTION_BEFORE,
  DIRECTION_AFTER
} DIRECTION;
static gint
emulator_get_text_length (struct vt_em *em);
static gint
emulator_get_text_offset_from_xy (struct vt_em *em,
			     gint x,
			     gint y);
static void
emulator_get_xy_from_text_offset (struct vt_em *em,
				  gint offset,
				  gint *x,
				  gint *y);
static gchar *
emulator_get_text_between_offsets (struct vt_em *em,
				   gint x1, gint y1,
				   gint x2, gint y2);
static struct vt_line *
emulator_get_line (struct vt_em *em,
		   gint line);
static gchar *
emulator_get_text (struct vt_em *em,
		   gint offset,
		   AtkTextBoundary boundary_type,
		   DIRECTION direction,
		   gint *x1, gint *y1,
		   gint *x2, gint *y2);
static void
zvt_accessible_class_init          (ZvtAccessibleClass *klass);
static void
zvt_accessible_finalize            (GObject            *object);

static void
zvt_accessible_real_initialize (AtkObject *obj,
                                gpointer  data);
static void
atk_text_interface_init (AtkTextIface *iface);
static gchar*
zvt_accessible_get_text                           (AtkText          *text,
						   gint             start_offset,
						   gint             end_offset);
static gunichar
zvt_accessible_get_character_at_offset            (AtkText          *text,
						   gint             offset);
static gchar*
zvt_accessible_get_text_after_offset              (AtkText          *text,
						   gint             offset,
						   AtkTextBoundary  boundary_type,
						   gint             *start_offset,
						   gint	            *end_offset);
static gchar*
zvt_accessible_get_text_at_offset                 (AtkText          *text,
						   gint             offset,
						   AtkTextBoundary  boundary_type,
						   gint             *start_offset,
						   gint             *end_offset);
static gchar*
zvt_accessible_get_text_before_offset             (AtkText          *text,
						   gint             offset,
						   AtkTextBoundary  boundary_type,
						   gint             *start_offset,
						   gint	            *end_offset);
static gint
zvt_accessible_get_caret_offset                   (AtkText          *text);
static void
zvt_accessible_get_character_extents              (AtkText          *text,
						   gint             offset,
						   gint             *x,
						   gint             *y,
						   gint             *width,
						   gint             *height,
						   AtkCoordType	    coords);
static AtkAttributeSet*
zvt_accessible_get_run_attributes              (AtkText	    *text,
						gint	  	    offset,
						gint             *start_offset,
						gint	 	    *end_offset);
static AtkAttributeSet*
zvt_accessible_get_default_attributes          (AtkText	    *text);
static gint
zvt_accessible_get_character_count                (AtkText          *text);
static gint
zvt_accessible_get_offset_at_point                (AtkText          *text,
						   gint             x,
						   gint             y,
						   AtkCoordType	    coords);
static gint
zvt_accessible_get_n_selections			  (AtkText          *text);
static gchar*
zvt_accessible_get_selection			  (AtkText          *text,
						   gint		    selection_num,
						   gint             *start_offset,
						   gint             *end_offset);
static gboolean
zvt_accessible_add_selection                      (AtkText          *text,
						   gint             start_offset,
						   gint             end_offset);
static gboolean
zvt_accessible_remove_selection                   (AtkText          *text,
						   gint		    selection_num);
static gboolean
zvt_accessible_set_selection                      (AtkText          *text,
						   gint		    selection_num,
						   gint             start_offset,
						   gint             end_offset);
static gboolean
zvt_accessible_set_caret_offset                   (AtkText          *text,
						   gint             offset);
static void
zvt_accessible_draw_text (void *user_data, struct vt_line *line, int row, int col, int len, int attr);
static   int  
zvt_accessible_cursor_state (void *user_data, int state);

typedef struct _ZvtAccessiblePriv ZvtAccessiblePriv;
struct _ZvtAccessiblePriv
{
  gint caret_position;

  /* Callback pointers */

  void (*real_draw_text)(void *user_data, struct vt_line *line, int row, int col, int len, int attr);
  void (*real_scroll_area)(void *user_data, int firstrow, int count, int offset, int fill);
  int  (*real_cursor_state)(void *user_data, int state);
};

static ZvtAccessiblePriv *zvt_accessible_get_private_data (ZvtAccessible *accessible);

static gpointer parent_class = NULL;
static GQuark quark_private_data = 0;

GType
zvt_accessible_get_type (void)
{
  static GType type = 0;

  if (!type)
  {
    static GTypeInfo tinfo =
    {
      0, /* class size */
      (GBaseInitFunc) NULL, /* base init */
      (GBaseFinalizeFunc) NULL, /* base finalize */
      (GClassInitFunc) zvt_accessible_class_init, /* class init */
      (GClassFinalizeFunc) NULL, /* class finalize */
      NULL, /* class data */
      0, /* instance size */
      0, /* nb preallocs */
      (GInstanceInitFunc) NULL, /* instance init */
      NULL /* value table */
    };

    static const GInterfaceInfo atk_text_info =
    {
        (GInterfaceInitFunc) atk_text_interface_init,
        (GInterfaceFinalizeFunc) NULL,
        NULL
    };

    /*
     * Figure out the size of the class and instance
     * we are deriving from
     */
    AtkObjectFactory *factory;
    GType derived_type;
    GTypeQuery query;
    GType derived_atk_type;

    derived_type = g_type_parent (ZVT_TYPE_TERM);
    factory = atk_registry_get_factory (atk_get_default_registry (), derived_type);
    derived_atk_type = atk_object_factory_get_accessible_type (factory);
    g_type_query (derived_atk_type, &query);
    tinfo.class_size = query.class_size;
    tinfo.instance_size = query.instance_size;

    type = g_type_register_static (derived_atk_type, "ZvtAccessible", &tinfo, 0);
    g_type_add_interface_static (type, ATK_TYPE_TEXT,
                                 &atk_text_info);
  }

  return type;
}

static void
zvt_accessible_real_initialize (AtkObject *obj,
                                gpointer  data)
{
  ZvtTerm *terminal;
  ZvtAccessible *zvt_accessible;
  ZvtAccessiblePriv *zvt_accessible_priv;

  ATK_OBJECT_CLASS (parent_class)->initialize (obj, data);

  zvt_accessible = ZVT_ACCESSIBLE (obj);
  terminal = ZVT_TERM (data);

  zvt_accessible_priv = zvt_accessible_get_private_data (zvt_accessible);

  zvt_accessible_priv->real_draw_text = terminal->vx->draw_text;
  terminal->vx->draw_text = zvt_accessible_draw_text;
  zvt_accessible_priv->real_cursor_state = terminal->vx->cursor_state;
  terminal->vx->cursor_state = zvt_accessible_cursor_state;

  zvt_accessible_priv->real_scroll_area = NULL;
}



AtkObject *
zvt_accessible_new(GtkWidget *widget)
{
  GObject *object;
  AtkObject *accessible;
  ZvtTerm *terminal;
  ZvtAccessible *zvt_accessible;
  ZvtAccessiblePriv *zvt_accessible_priv;

  object = g_object_new (ZVT_TYPE_ACCESSIBLE, NULL);

  accessible = ATK_OBJECT (object);
  atk_object_initialize (accessible, widget);
  accessible->role = ATK_ROLE_TERMINAL;

  return accessible;
}

ZvtAccessiblePriv *
zvt_accessible_get_private_data (ZvtAccessible *accessible)
{
  ZvtAccessiblePriv *private_data;

  private_data = g_object_get_qdata (G_OBJECT (accessible),
                                     quark_private_data);
  if (!private_data)
    {
      private_data = g_new (ZvtAccessiblePriv, 1);
      g_object_set_qdata (G_OBJECT (accessible),
                          quark_private_data,
                          private_data);
    }
  return private_data;
}

static gint
emulator_get_text_length (struct vt_em *em)
{
  struct vt_line *current = (struct vt_line *) em->lines.head;
  gint length = 0;

  while (current)
  {
    length += current->width+1; /* Account for phantom linefeed at end of line */
    current = current->next;
  }
  return length;
}



static gint
emulator_get_text_offset_from_xy (struct vt_em *em,
			     gint x,
			     gint y)
{
  struct vt_line *current = (struct vt_line *) em->lines.head;
  gint offset = 0;

  while (current)
    {
      if (current->line > x)
	break;
      else if (current->line == x)
	offset += y;
      else
	offset += current->width+1;
      current = current->next;
  }
  return offset;
}



static void
emulator_get_xy_from_text_offset (struct vt_em *em,
				  gint offset,
				  gint *x,
				  gint *y)
{
  struct vt_line *first = (struct vt_line *) em->lines.head;
  struct vt_line *current = first;
  gint current_offset = 0;

  *x = *y = 0;
  while (current && current_offset < offset)
    {
      if (offset-current_offset > current->width)
	{
	  (*x)++;
	  current_offset += current->width+1;
	}
      else
	{
	  *y += (current_offset-offset);
	  if (*y >= current->width)
	    {
	      *x = *y = -1;
	    }
	  current_offset = offset;
	}
      current = current->next;
    }
  return;
}



static gchar *
emulator_get_text_between_offsets (struct vt_em *em,
				   gint x1, gint y1,
				   gint x2, gint y2)
{
  struct vt_line *first = (struct vt_line *) em->lines.head;
  struct vt_line *current = first;
  gint length = 0;
  gchar *text, *p;
  gint i;
  
  /* Ensure x1, y1 < x2, y2 */
  if (x1 > x2)
    {
      i = x1;
      x1 = x2;
      x2 = i;
      i = y1;
      y1 = y2;
      y2 = i;
    }
  else if (x1 == x2 && y1 > y2)
    {
      i = y1;
      y1 = y2;
      y2 = i;
    }

  while (current)
    {
      if (current->line > y2)
	break;
      if (current->line == y1)
	length += 1+(current->width)-x1;
      else if (current->line == y2)
	length += x2+1;
      else if (current->line > y1 && current->line < y2)
	length += current->width+1;
    }

  text = p = g_new (gchar, length);
  current = first;
  while (current)
    {
      gint *data;
      gint bytesToCopy;
      if (current->line > y2)
	break;
      if (current->line == y1)
	{
	  data = &(current->data[x1]);
	  bytesToCopy = current->width-x1;
	  if (bytesToCopy < 0)
	    bytesToCopy = 0;
	}
      else if (current->line > y1 && current->line < y2)
	{
	  data = current->data;
	  bytesToCopy = current->width;
	}
      else
	{
	  data = current->data;
	  bytesToCopy = (x2 > current->width) ? current->width : x2;
	}
      while (bytesToCopy--)
	*p++ = (gchar) (*data++) & VTATTR_DATAMASK;
      if (current->line >= y1 && current->line <= y2)
	*p++ = '\n';
      current = current->next;
    }
  return text;
}



static struct vt_line *
emulator_get_line (struct vt_em *em,
		   gint line)
{
  struct vt_line *current = (struct vt_line *) em->lines.head;
  
  while (current)
    {
      if (current->line == line)
	break;
      current = current->next;
    }
  if (current->line == line)
    return current;
  else
    return NULL;
}



static gchar *
emulator_get_text (struct vt_em *em,
		   gint offset,
		   AtkTextBoundary boundary_type,
		   DIRECTION direction,
		   gint *x1, gint *y1,
		   gint *x2, gint *y2)
{
  gint x, y;
  struct vt_line *current_line;

  emulator_get_xy_from_text_offset (em,
				  offset, &x, &y);
  current_line = emulator_get_line (em, y);

  switch (boundary_type)
    {
    case ATK_TEXT_BOUNDARY_CHAR:
      if (direction == DIRECTION_BEFORE)
	{
	  *x1 = x-1;
	  if (*x1 < 0)
	    return g_strdup ("");
	  *x2 = x;
	  *y1 = *y2 = y;
	}
      if (direction == DIRECTION_AT)
	{
	  *x1 = x;
	  *x2 = x+1;
	  *y1 = *y2 = y;
	}
      if (direction == DIRECTION_AFTER)
	{
	  *x1 = x+1;
	  if (*x1 >= current_line->width)
	    return g_strdup ("");
	  *x2 = x+2;
	  *y1 = *y2 = y;
	}
    case ATK_TEXT_BOUNDARY_LINE_START:
    case ATK_TEXT_BOUNDARY_LINE_END:
      *x1 = 0;
      if (direction == DIRECTION_BEFORE)
	current_line = current_line->prev;
      else if (direction == DIRECTION_AFTER)
	current_line = current_line->next;
      if (!current_line)
	return g_strdup ("");
      *y1 = *y2 = current_line->line;
      *x2 = current_line->width-1;
    default:
      break;
}
  return emulator_get_text_between_offsets (em,
					  *x1, *y1, *x2, *y2);
}



static void
zvt_accessible_class_init (ZvtAccessibleClass *klass)
{
  AtkObjectClass *class = ATK_OBJECT_CLASS (klass);
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

  parent_class = g_type_class_peek_parent (klass);

  class->initialize = zvt_accessible_real_initialize;

  gobject_class->finalize = zvt_accessible_finalize;

  quark_private_data = g_quark_from_static_string ("zvt-accessible-private-data");
}

static void
zvt_accessible_finalize (GObject *object)
{
  ZvtAccessible *accessible;
  ZvtAccessiblePriv *accessible_priv;

  accessible = ZVT_ACCESSIBLE (object);
  accessible_priv = zvt_accessible_get_private_data (accessible);
  if (accessible_priv)
    {
      g_free (accessible_priv);
      g_object_set_qdata (G_OBJECT (accessible),
                          quark_private_data,
                          NULL);
    }
  G_OBJECT_CLASS (parent_class)->finalize (object); 
}

static void
atk_text_interface_init (AtkTextIface *iface)
{
  g_return_if_fail (iface != NULL);
  iface->get_text = zvt_accessible_get_text;
  iface->get_character_at_offset = zvt_accessible_get_character_at_offset;
  iface->get_text_before_offset = zvt_accessible_get_text_before_offset;
  iface->get_text_at_offset = zvt_accessible_get_text_at_offset;
  iface->get_text_after_offset = zvt_accessible_get_text_after_offset;
  iface->get_character_count = zvt_accessible_get_character_count;
  iface->get_caret_offset = zvt_accessible_get_caret_offset;
  iface->set_caret_offset = zvt_accessible_set_caret_offset;
  iface->get_n_selections = zvt_accessible_get_n_selections;
  iface->get_selection = zvt_accessible_get_selection;
  iface->add_selection = zvt_accessible_add_selection;
  iface->remove_selection = zvt_accessible_remove_selection;
  iface->set_selection = zvt_accessible_set_selection;
  iface->get_character_extents = zvt_accessible_get_character_extents;
  iface->get_offset_at_point = zvt_accessible_get_offset_at_point;
  iface->get_run_attributes = zvt_accessible_get_run_attributes;
  iface->get_default_attributes = zvt_accessible_get_default_attributes;
}



static gchar*
zvt_accessible_get_text                           (AtkText          *text,
						   gint             start_offset,
						   gint             end_offset)
{
  GtkWidget *widget;
  ZvtTerm *terminal;
  struct vt_em *emulator;
  gint x1, y1, x2, y2;

  g_return_val_if_fail (GTK_IS_ACCESSIBLE(text), NULL);
  widget = GTK_ACCESSIBLE (text)->widget;
  if (widget == NULL)
  {
    /* State is defunct */
    return NULL;
  }
  g_return_val_if_fail (ZVT_IS_TERM(widget), NULL);
  terminal = ZVT_TERM(widget);
  emulator = &(terminal->vx->vt);
  emulator_get_xy_from_text_offset (emulator,
				    start_offset, &x1, &y1);
  emulator_get_xy_from_text_offset (emulator,
				    end_offset, &x2, &y2);
  return emulator_get_text_between_offsets (emulator,
					   x1, y1, x2, y2);
}



static gunichar
zvt_accessible_get_character_at_offset            (AtkText          *text,
						   gint             offset)
{
  GtkWidget *widget;
  ZvtTerm *terminal;
  struct vt_em *emulator;
  gint x, y;
  gchar *character;
  gunichar rv;

  g_return_val_if_fail (GTK_IS_ACCESSIBLE(text), (gunichar)0);
  widget = GTK_ACCESSIBLE (text)->widget;
  if (widget == NULL)
  {
    /* State is defunct */
    return (gunichar) 0;
  }
  g_return_val_if_fail (ZVT_IS_TERM(widget), (gunichar) 0);
  terminal = ZVT_TERM(widget);
  emulator = &(terminal->vx->vt);
  emulator_get_xy_from_text_offset (emulator, offset,
				    &x, &y);
  character = emulator_get_text_between_offsets (emulator,
						x, y, x, y);
  rv = *character;
  g_free (character);
  return rv;
}



static gchar*
zvt_accessible_get_text_after_offset              (AtkText          *text,
						   gint             offset,
						   AtkTextBoundary  boundary_type,
						   gint             *start_offset,
						   gint	            *end_offset)
{
  GtkWidget *widget;
  ZvtTerm *terminal;
  struct vt_em *emulator;
  gint x1, y1, x2, y2;
  gchar *rv;

  g_return_val_if_fail (GTK_IS_ACCESSIBLE(text), NULL);
  widget = GTK_ACCESSIBLE (text)->widget;
  if (widget == NULL)
  {
    /* State is defunct */
    return NULL;
  }
  g_return_val_if_fail (ZVT_IS_TERM(widget), NULL);
  terminal = ZVT_TERM(widget);
  emulator = &(terminal->vx->vt);
  rv = emulator_get_text (emulator,
			    offset,
			    boundary_type, DIRECTION_AFTER,
			    &x1, &y1, &x2, &y2);
  *start_offset = emulator_get_text_offset_from_xy (emulator,
						  x1, y1);
  *end_offset = emulator_get_text_offset_from_xy (emulator,
						  x2, y2);
  return rv;
}



static gchar*
zvt_accessible_get_text_at_offset                 (AtkText          *text,
						   gint             offset,
						   AtkTextBoundary  boundary_type,
						   gint             *start_offset,
						   gint             *end_offset)
{
  GtkWidget *widget;
  ZvtTerm *terminal;
  struct vt_em *emulator;
  gint x1, y1, x2, y2;
  gchar *rv;

  g_return_val_if_fail (GTK_IS_ACCESSIBLE(text), NULL);
  widget = GTK_ACCESSIBLE (text)->widget;
  if (widget == NULL)
  {
    /* State is defunct */
    return NULL;
  }
  g_return_val_if_fail (ZVT_IS_TERM(widget), NULL);
  terminal = ZVT_TERM(widget);
  emulator = &(terminal->vx->vt);
  rv = emulator_get_text (emulator,
			    offset,
			    boundary_type, DIRECTION_AT,
			    &x1, &y1, &x2, &y2);
  *start_offset = emulator_get_text_offset_from_xy (emulator,
						  x1, y1);
  *end_offset = emulator_get_text_offset_from_xy (emulator,
						  x2, y2);
  return rv;
}



static gchar*
zvt_accessible_get_text_before_offset             (AtkText          *text,
						   gint             offset,
						   AtkTextBoundary  boundary_type,
						   gint             *start_offset,
						   gint	            *end_offset)
{
  GtkWidget *widget;
  ZvtTerm *terminal;
  struct vt_em *emulator;
  gint x1, y1, x2, y2;
  gchar *rv;

  g_return_val_if_fail (GTK_IS_ACCESSIBLE(text), NULL);
  widget = GTK_ACCESSIBLE (text)->widget;
  if (widget == NULL)
  {
    /* State is defunct */
    return NULL;
  }
  g_return_val_if_fail (ZVT_IS_TERM(widget), NULL);
  terminal = ZVT_TERM(widget);
  emulator = &(terminal->vx->vt);
  rv = emulator_get_text (emulator,
			    offset,
			    boundary_type, DIRECTION_BEFORE,
			    &x1, &y1, &x2, &y2);
  *start_offset = emulator_get_text_offset_from_xy (emulator,
						  x1, y1);
  *end_offset = emulator_get_text_offset_from_xy (emulator,
						  x2, y2);
  return rv;
}



static gint
zvt_accessible_get_caret_offset                   (AtkText          *text)
{
  ZvtAccessible *acc;
  ZvtAccessiblePriv *acc_priv;
  g_return_val_if_fail (ZVT_IS_ACCESSIBLE(text), -1);
  acc = ZVT_ACCESSIBLE(text);
  acc_priv = zvt_accessible_get_private_data (acc);
  return acc_priv->caret_position;
}



static void
zvt_accessible_get_character_extents              (AtkText          *text,
						   gint             offset,
						   gint             *x,
						   gint             *y,
						   gint             *width,
						   gint             *height,
						   AtkCoordType	    coords)
{
  GtkWidget *widget;
  ZvtTerm *terminal;
  struct vt_em *emulator;
  gint window_x, window_y;

  g_return_if_fail (GTK_IS_ACCESSIBLE(text));
  widget = GTK_ACCESSIBLE (text)->widget;
  if (widget == NULL)
  {
    /* State is defunct */
    return;
  }
  g_return_if_fail (ZVT_IS_TERM(widget));
  terminal = ZVT_TERM(widget);
  emulator = &(terminal->vx->vt);
  emulator_get_xy_from_text_offset (emulator,
				    offset, x, y);
  *x *= terminal->charwidth;
  *y = *y * terminal->charheight - emulator->scrollbackoffset;
  if (coords == ATK_XY_SCREEN)
    {
      gdk_window_get_position (terminal->term_window, &window_x, &window_y);
      *x += window_x;
      *y += window_y;
    }
  *width = terminal->charwidth;
  *height = terminal->charheight;
}



static AtkAttributeSet*
zvt_accessible_get_run_attributes              (AtkText	    *text,
						gint	  	    offset,
						gint             *start_offset,
						gint	 	    *end_offset)
{
  return NULL;
}



static AtkAttributeSet*
zvt_accessible_get_default_attributes          (AtkText	    *text)
{
  return NULL;
}



static gint
zvt_accessible_get_character_count                (AtkText          *text)
{
  GtkWidget *widget;
  ZvtTerm *terminal;
  struct vt_em *emulator;

  g_return_val_if_fail (GTK_IS_ACCESSIBLE(text), -1);
  widget = GTK_ACCESSIBLE (text)->widget;
  if (widget == NULL)
  {
    /* State is defunct */
    return -1;
  }
  g_return_val_if_fail (ZVT_IS_TERM(widget), -1);
  terminal = ZVT_TERM(widget);
  emulator = &(terminal->vx->vt);
  return emulator_get_text_length (emulator);
}



static gint
zvt_accessible_get_offset_at_point                (AtkText          *text,
						   gint             x,
						   gint             y,
						   AtkCoordType	    coords)
{
  GtkWidget *widget;
  ZvtTerm *terminal;
  struct vt_em *emulator;
  gint local_x, local_y;

  g_return_val_if_fail (GTK_IS_ACCESSIBLE(text), -1);
  widget = GTK_ACCESSIBLE (text)->widget;
  if (widget == NULL)
  {
    /* State is defunct */
    return -1;
  }

  g_return_val_if_fail (ZVT_IS_TERM(widget), -1);
  terminal = ZVT_TERM(widget);
  emulator = &(terminal->vx->vt);
  local_x = x/terminal->charwidth;
  local_y = y / terminal->charheight + emulator->scrollbackoffset;
  return emulator_get_text_offset_from_xy (emulator,
					   local_x, local_y);
}




static gint
zvt_accessible_get_n_selections			  (AtkText          *text)
{
  GtkWidget *widget;
  ZvtTerm *terminal;

  g_return_val_if_fail (GTK_IS_ACCESSIBLE(text), -1);
  widget = GTK_ACCESSIBLE (text)->widget;
  if (widget == NULL)
  {
    /* State is defunct */
    return -1;
  }

  g_return_val_if_fail (ZVT_IS_TERM(widget), -1);
  terminal = ZVT_TERM(widget);
  return (terminal->vx->selected) ? 1 : 0;
}



static gchar*
zvt_accessible_get_selection			  (AtkText          *text,
						   gint		    selection_num,
						   gint             *start_offset,
						   gint             *end_offset)
{
  GtkWidget *widget;
  ZvtTerm *terminal;
  struct vt_em *emulator;
  
  g_return_val_if_fail (GTK_IS_ACCESSIBLE(text), NULL);
  widget = GTK_ACCESSIBLE (text)->widget;
  if (widget == NULL)
  {
    /* State is defunct */
    return NULL;
  }

  g_return_val_if_fail (ZVT_IS_TERM(widget), NULL);
  terminal = ZVT_TERM(widget);
  if (!terminal->vx->selected || selection_num != 0)
    {
      *start_offset = *end_offset = -1;
      return NULL;
    }
emulator = &(terminal->vx->vt);
  *start_offset = emulator_get_text_offset_from_xy (emulator,
						    terminal->vx->selstartx, terminal->vx->selstarty);
  *end_offset = emulator_get_text_offset_from_xy (emulator,
						  terminal->vx->selendx, terminal->vx->selendy);
  return emulator_get_text_between_offsets (emulator,
					    terminal->vx->selstartx, terminal->vx->selstarty,
					    terminal->vx->selendx, terminal->vx->selendy);
}



static gboolean
zvt_accessible_add_selection                      (AtkText          *text,
						   gint             start_offset,
						   gint             end_offset)
{
  GtkWidget *widget;
  ZvtTerm *terminal;
  struct vt_em *emulator;

  g_return_val_if_fail (GTK_IS_ACCESSIBLE(text), FALSE);
  widget = GTK_ACCESSIBLE (text)->widget;
  if (widget == NULL)
  {
    /* State is defunct */
    return FALSE;
  }

  g_return_val_if_fail (ZVT_IS_TERM(widget), FALSE);
  terminal = ZVT_TERM(widget);
  if (terminal->vx->selected)
    return FALSE;
  emulator = &(terminal->vx->vt);
  terminal->vx->selected = 1;
  emulator_get_xy_from_text_offset (emulator,
				   start_offset, &(terminal->vx->selstartx), &(terminal->vx->selstarty));
  emulator_get_xy_from_text_offset (emulator,
				  end_offset, &(terminal->vx->selendx), &(terminal->vx->selendy));
  return TRUE;
}



static gboolean
zvt_accessible_remove_selection                   (AtkText          *text,
						   gint		    selection_num)
{
  GtkWidget *widget;
  ZvtTerm *terminal;

  g_return_val_if_fail (GTK_IS_ACCESSIBLE(text), FALSE);
  widget = GTK_ACCESSIBLE (text)->widget;
  if (widget == NULL)
  {
    /* State is defunct */
    return FALSE;
  }

  g_return_val_if_fail (ZVT_IS_TERM(widget), FALSE);
  terminal = ZVT_TERM(widget);
  if (!terminal->vx->selected || selection_num != 0)
    return FALSE;
  terminal->vx->selected = 0;
  return TRUE;
}



static gboolean
zvt_accessible_set_selection                      (AtkText          *text,
						   gint		    selection_num,
						   gint             start_offset,
						   gint             end_offset)
{
  GtkWidget *widget;
  ZvtTerm *terminal;
  struct vt_em *emulator;

  g_return_val_if_fail (GTK_IS_ACCESSIBLE(text), FALSE);
  widget = GTK_ACCESSIBLE (text)->widget;
  if (widget == NULL)
  {
    /* State is defunct */
    return FALSE;
  }

  g_return_val_if_fail (ZVT_IS_TERM(widget), FALSE);
  terminal = ZVT_TERM(widget);
  if (selection_num != 0)
    return FALSE;
  emulator = &(terminal->vx->vt);
  terminal->vx->selected = 1;
  emulator_get_xy_from_text_offset (emulator,
				   start_offset, &(terminal->vx->selstartx), &(terminal->vx->selstarty));
  emulator_get_xy_from_text_offset (emulator,
				  end_offset, &(terminal->vx->selendx), &(terminal->vx->selendy));
  return TRUE;
}



static gboolean
zvt_accessible_set_caret_offset                   (AtkText          *text,
						   gint             offset)
{
  ZvtAccessible *acc;
  ZvtAccessiblePriv *acc_priv;
  g_return_val_if_fail (ZVT_IS_ACCESSIBLE(text), -1);
  acc = ZVT_ACCESSIBLE(text);
  acc_priv = zvt_accessible_get_private_data (acc);
  acc_priv->caret_position = offset;
  return TRUE;
}



static void
zvt_accessible_draw_text (void *user_data, struct vt_line *line, int row, int col, int len, int attr)
{
  GtkWidget *widget;
  AtkObject *accessible;
  ZvtAccessible *zvt_accessible;
  ZvtAccessiblePriv *zvt_accessible_priv;
  ZvtTerm *terminal;
  gint offset;

  widget = user_data;
  g_return_if_fail (GTK_IS_WIDGET(widget));
  g_return_if_fail (ZVT_IS_TERM(widget));
  terminal = ZVT_TERM (widget);
  accessible = gtk_widget_get_accessible (widget);
  g_return_if_fail (ZVT_IS_ACCESSIBLE(accessible));
  zvt_accessible = ZVT_ACCESSIBLE(accessible);
  zvt_accessible_priv = zvt_accessible_get_private_data (zvt_accessible);
  if (zvt_accessible_priv->real_draw_text == NULL)
    {
      /* We're hosed!!! */
      return;
    }
  zvt_accessible_priv->real_draw_text (user_data,
				  line, row, col, len, attr);
  
  offset = emulator_get_text_offset_from_xy (&terminal->vx->vt,
					     col, row);
  g_signal_emit_by_name (accessible, "text_changed::insert",
                         offset, len);
}



static   int  
zvt_accessible_cursor_state (void *user_data, int state)
{
  GtkWidget *widget;
  ZvtTerm *term;
  AtkObject *accessible;
  ZvtAccessible *zvt_accessible;
  ZvtAccessiblePriv *zvt_accessible_priv;
  gint offset;

  widget = user_data;
  g_return_val_if_fail (GTK_IS_WIDGET(widget), 0);
  g_return_val_if_fail (ZVT_IS_TERM(widget), 0);
  term = ZVT_TERM(widget);
  accessible = gtk_widget_get_accessible (widget);
  g_return_val_if_fail (ZVT_IS_ACCESSIBLE(accessible), 0);
  zvt_accessible = ZVT_ACCESSIBLE(accessible);
  offset = emulator_get_text_offset_from_xy (&term->vx->vt,
					      term->vx->vt.cursorx, term->vx->vt.cursory);
  zvt_accessible_priv = zvt_accessible_get_private_data (zvt_accessible);
  if (offset != zvt_accessible_priv->caret_position)
    {
      zvt_accessible_priv->caret_position = offset;
      g_signal_emit_by_name (accessible, "text_caret_moved",
			     offset);
    }
  if (zvt_accessible_priv->real_cursor_state)
    return zvt_accessible_priv->real_cursor_state (user_data, state);
  return 0;
}



